from __future__ import annotations
from typing import Protocol

import numpy as np
import numpy.typing as npt

from .. import numpy as xtnp
from .. import tensor as xtt

'''
Args functions:
    For reduction functions whose return values are
    **indices** or **coordinates** over multiple axes
    E.g. argsmax
'''


ARGS_DIM = 'ARGS_DIM'

class _args_func(Protocol):
    def __call__(self, a: np.ndarray, axes: xtnp.AxesLike) -> npt.NDArray[np.int_]: ...

class ArgsFunction(Protocol):
    def __call__(self, x: xtt.TensorLike, /, dims: xtt.DimsLike) -> xtt.XTensor: ...

class CoordsFunction(Protocol):
    def __call__(self, x: xtt.TensorLike, /, dims: xtt.DimsLike, *, 
            use_index_if_no_coord: bool=False) -> xtt.XTensor: ...


def _reduction_factory(_func: _args_func) -> ArgsFunction:
    @xtt.generalize_at_0
    def _reduce(X: xtt.XTensor, /, dims: xtt.DimsLike) -> xtt.XTensor:
        '''
        Return
        '''

        axes = X.get_axes(dims)
        r_dims, s_dims = xtt.strip(X.dims, axes, only_remaining=False)
        r_coords, s_coords = xtt.strip(X.coords, axes, only_remaining=False)

        new_dims = r_dims + [ARGS_DIM]
        new_coords = r_coords + [[dim for dim in s_dims]]

        args = _func(X.data, axes=axes)
        return xtt.XTensor(args, dims=new_dims, coords=new_coords)
    return _reduce


argsmin = _reduction_factory(xtnp.argsmin)
argsmax = _reduction_factory(xtnp.argsmax)
nanargsmin = _reduction_factory(xtnp.nanargsmin)
nanargsmax = _reduction_factory(xtnp.nanargsmax)



def _coord_reduc_factory(_func: ArgsFunction) -> CoordsFunction:
    @xtt.generalize_at_0
    def _reduce(X: xtt.XTensor, /, dims: xtt.DimsLike, *, use_index_if_no_coord: bool=False) -> xtt.XTensor:
        '''
        Return the coordinate on [dim] that maximizes/minimizes x If dimension [dim] does
        not have coordinates, this is equivalent to (nan)argmax/argmin
        '''
        args = _func(X, dims)
        axes = X.get_axes(dims)

        coords = [X.coords[axis] for axis in axes]
        coords_r = []

        for i, (axis, coord) in enumerate(zip(axes, coords)):

            if coord is None: 
                if use_index_if_no_coord:
                    coord_ = np.arange(X.shape[axis])
                else:
                    raise ValueError(f'Tensor has no coordinates on axis {axis}')
            else:
                coord_ = coord


            coords_r.append(coord_[args.data[...,i]])

        return xtt.XTensor(
                np.stack(coords_r, axis=-1),
                dims=args.dims, 
                coords=args.coords)

    return _reduce


coordsmin = _coord_reduc_factory(argsmin)
coordsmax = _coord_reduc_factory(argsmax)
nancoordsmin = _coord_reduc_factory(nanargsmin)
nancoordsmax = _coord_reduc_factory(nanargsmax)





from __future__ import annotations
from numbers import Real
from typing import Any, Callable, Generic, List, Protocol, Sequence, Tuple, TypeVar, Union
from typing_extensions import ParamSpec, TypeVarTuple

import numpy as np
import numpy.typing as npt

T_con = TypeVar('T_con', contravariant=True)
U_con = TypeVar('U_con', contravariant=True)
V_con = TypeVar('V_con', contravariant=True)
T_co = TypeVar('T_co', covariant=True)

S = TypeVar('S')
T = TypeVar('T')
O = ParamSpec('O')

class Array(Protocol):
    def __array__(self) -> np.ndarray: ...


TensorLike = Union[Array, List, float]

AxesPermutation = List[Union[int, None]]
'''
A list of integers or None that represents a permutation of a set of axes. None
means that a new axis of length 1 is to be created at the corresponding index.
'''

DimLike = Union[str,int,Tuple[str,int]]
DimsLike = List[DimLike]

Dims = List[Union[str,None]]
Coords = List[Union[npt.NDArray[Any],None]]


class Function_1Arg(Protocol[T_con, O, T_co]):
    def __call__(self, X: T_con, /, *args: O.args, **kwargs: O.kwargs) -> T_co: ...


class Function_2Args(Protocol[T_con, U_con, O, T_co]):
    def __call__(self, X: T_con, Y: U_con, /, *args: O.args, **kwargs: O.kwargs) -> T_co: ...


class Function_3Args(Protocol[T_con, U_con, V_con, O, T_co]):
    def __call__(self, X: T_con, Y: U_con, Z: V_con, /, *args: O.args, **kwargs: O.kwargs) -> T_co: ...


class BinaryOperator(Protocol[T]):
    def __call__(self, X: T, Y: T) -> T: ...


class TernaryOperator(Protocol[T]):
    def __call__(self, X: T, Y: T, Z: T) -> T: ...














from ._base import to_xtensor

from ._axes import permute, newdims, align, shapes_broadcastable, permutation_well_defined, shape, stack, rank

from ._dims import mergedims, flatten, dimsfirst, dimslast, name_dim_if_absent, dims

from ._coords import mergecoords, coords_same

from ._generalize import generalize_at_0, generalize_at_1, generalize_at_2, generalize_at_3


from ._misc import strip 




from __future__ import annotations
from typing import Union

import numpy as np

from numpy import typing as nptyping

import xarray as xr
import torch



NDArray = Union[nptyping.NDArray[np.float_], nptyping.NDArray[np.int_], nptyping.NDArray[np.bool_], xr.DataArray, torch.Tensor]

number = Union[int, float, complex, np.int_, np.float_, np.complex_, np.bool_]
Real: Union[int, float, np.int_, np.float_]





from __future__ import annotations
from typing import Tuple, Union




AxisDimPair = Tuple[int, str]
DimLike = Union[int, str, AxisDimPair]
DimsLike = Union[Tuple[int,...], Tuple[str,...], Tuple[AxisDimPair, ...]]






import numpy as np
from .. import numpy as xtnp
from .. import tensor as xtt


def confusion_matrix(truth: np.ndarray, pred: np.ndarray, /, *, n_classes: int) -> np.ndarray:
    '''
        X: (*, M)
        Y: (*, M)
    '''
    z = truth* n_classes + pred
    cmat = xtnp.bincount(z, N=n_classes**2)
    return cmat.reshape(*cmat.shape[:-1], n_classes, n_classes)


def get_confmat_function(target_dim: str, truth_dim: str, pred_dim: str, n_classes: int):
    @xtt.generalize_at_0
    @xtt.generalize_at_1
    def wrapped_confmat(truth: xtt.XTensor, pred: xtt.XTensor) -> xtt.XTensor:
        
        X = xtt.name_dim_if_absent(truth, -1, target_dim)
        Y = xtt.name_dim_if_absent(pred, -1, target_dim)

        X = xtt.dimslast(truth, [target_dim])
        Y = xtt.dimslast(pred, [target_dim])

        x, y, dims, coords = xtt.vanilla_broadcaster(X, Y)
        cm = confusion_matrix(x, y, n_classes=n_classes)

        C = xtt.XTensor(cm,
                dims=dims[:-1] + [truth_dim, pred_dim],
                coords=coords[:-1] + [None, None]
            )
        return C

    return wrapped_confmat




from __future__ import annotations
from typing import TYPE_CHECKING, List, Tuple



from ..typing import AxesPermutation

if TYPE_CHECKING:
    from .._base import XTensor

from ._types import Dimcaster


def unilateral_dimcast(strict: bool=False) -> Dimcaster:
    def _dimcast(X: XTensor, Y: XTensor) -> Tuple[AxesPermutation, AxesPermutation]:
        axesp = castdim(X, Y, strict=strict)
        return [i for i in range(X.rank)], axesp
    return _dimcast

def trivial_dimcast(X: XTensor, Y: XTensor):
    axes_x: List[int|None] = [i for i in range(X.rank)]
    axes_y: List[int|None] = [i for i in range(Y.rank)]
    return axes_x, axes_y


def castdim(target: XTensor, subject: XTensor, strict: bool=False) -> AxesPermutation:
    '''
    Cast the dimensions of subject so that the two tensors are ready to be
    broadcast together.
    '''

    dims_t = list(target.dims)
    dims_s = list(subject.dims)
    max_unmatched_named_dimension_migration_index = len(subject.dims) - len(target.dims)

    # pad unnamed singletons so that the target tensor does not have fewer
    # dimensions than the subject tensor
    if len(dims_t) < len(dims_s):
        if strict:
            raise ValueError(f'Dimcast impossible with dimensions {dims_t} {dims_s} with strict=True')
        dims_t = [None for _ in range(len(dims_t), len(dims_s))] + dims_t
    # else:
    #     dims_s = [None for _ in range(len(dims_s), len(dims_t))] + dims_s

    ts_map = dict()

    # match named dimensions
    for axis_t, dim_t in enumerate(dims_t):
        if dim_t is not None:
            if dim_t in dims_s:
                ts_map[axis_t] = subject.get_axis(dim_t)

    axes_s_not_mached = [axis for axis in range(len(dims_s)) if axis not in ts_map.values()]


    dimcast_possible = True

    axes = []
    try:
        # match remaning
        for axis_t, dim_t in enumerate(dims_t):
            if axis_t not in ts_map.keys():
                # axis in target tensor is not yet matched
                axis_s = axis_t - len(dims_t) + len(dims_s)
                if axis_s >= 0:
                    dim_s = dims_s[axis_s]
                    if axis_s not in ts_map.values():
                        # axis in subject tensor is not yet matched
                        if dim_t is None or dim_s is None:
                            ts_map[axis_t] = axis_s
                        else:
                            raise ValueError()
                    else:
                        # axis in subject tensor is already matched
                        if axis_t < max_unmatched_named_dimension_migration_index:
                            axis_s = axes_s_not_mached.pop(0)
                            if dims_s[axis_s] is not None:
                                ts_map[axis_t] = axis_s
                            else:
                                raise ValueError()
                        else:
                            ts_map[axis_t] = None
                else:
                    ts_map[axis_t] = None
                    
        axes = [ts_map[axis_t] for axis_t in range(len(dims_t))]

        for axis in range(len(subject.shape)):
            if axis not in axes: raise ValueError()

    except ValueError:
        dimcast_possible = False 
    
    if not dimcast_possible:
        raise ValueError(f'Dimcast impossible with dimensions {dims_t} {dims_s}')

    return axes






from __future__ import annotations
from typing import Protocol, Tuple
import numpy as np

from .. import tensor as xtt


'''
For reduction functions that can act over multiple axes/dims.
'''

class _np_reduction_func(Protocol):
    def __call__(self, a: np.ndarray, axis: int|Tuple[int,...]) -> np.ndarray: ...


class ReductionFunc(Protocol):
    def __call__(self, x: xtt.TensorLike,/, dim: xtt.DimLike|xtt.DimsLike|None) -> xtt.XTensor: ...


def _reduction_factory(_np_func: _np_reduction_func) -> ReductionFunc:
    @xtt.generalize_at_0
    def _reduce(X: xtt.XTensor, /, dim: xtt.DimLike|xtt.DimsLike|None=None) -> xtt.XTensor:
        axes = X.get_axes(dim)
        _y = _np_func(X.data, axis=tuple(axes))

        return xtt.XTensor(_y, dims=xtt.strip(X.dims, axes), coords=xtt.strip(X.coords, axes))
    return _reduce


_sum = _reduction_factory(np.sum)
_mean = _reduction_factory(np.mean)
_std = _reduction_factory(np.std)

_nanmean = _reduction_factory(np.nanmean)
_nanstd = _reduction_factory(np.nanstd)
_nansum = _reduction_factory(np.nansum)

_max = _reduction_factory(np.max)
_min = _reduction_factory(np.min)

_nanmax = _reduction_factory(np.nanmax)
_nanmin = _reduction_factory(np.nanmin)

_all = _reduction_factory(np.all)
_any = _reduction_factory(np.any)





from __future__ import annotations
from functools import wraps
import numpy.typing as npt
from typing import TYPE_CHECKING, Callable, Tuple, TypeVar 
from typing_extensions import ParamSpec, Concatenate


from .typing import BinaryOperator, Dims, Coords, TernaryOperator

from .broadcast._types import Broadcaster

from .broadcast._broadcast import vanilla_broadcaster, cast

from .basic_utils import mergedims, mergecoords

if TYPE_CHECKING:
    from ._base import XTensor


O = ParamSpec('O')
T = TypeVar('T')


def promote_binary_operator(
        broadcaster: Broadcaster|None=None, 
        # dimcoord_converter: Callable[[Dims, Coords], Tuple[Dims, Coords]]|None=None,
    ) -> Callable[
            [BinaryOperator[npt.NDArray]], BinaryOperator[XTensor]
        ]:
    '''
    Promote an NDArray binary operator to XTensor binary operator
    '''
    if broadcaster is None: broadcaster = vanilla_broadcaster
    def wrapper(f: BinaryOperator[npt.NDArray]) -> BinaryOperator[XTensor]:
        @wraps(f)
        def wrapped(X: XTensor, Y: XTensor) -> XTensor:
            from ._base import XTensor
            
            _x, _y, dims, coords = broadcaster(X, Y)
            res_data = f(_x, _y)

            # if dimcoord_converter:
            #     dims, coords = dimcoord_converter(dims, coords)
            return XTensor(res_data, dims, coords)
        return wrapped
    return wrapper


def promote_ternary_operator(
    ) -> Callable[[TernaryOperator[npt.NDArray]], TernaryOperator[XTensor]]:
    '''
    Promote an NDArray ternary operator to XTensor ternary operator using
    vanilla broadcaster
    '''
    broadcaster = vanilla_broadcaster
    def wrapper(f: TernaryOperator[npt.NDArray]) -> TernaryOperator[XTensor]:
        @wraps(f)
        def wrapped(X: XTensor, Y: XTensor, Z: XTensor) -> XTensor:
            from ._base import XTensor

            largest = X
            if X.rank < Y.rank: largest = Y
            if largest.rank < Z.rank: largest = Z

            _cast = cast(broadcaster, largest)

            X1 = _cast(X)
            Y1 = _cast(Y)
            Z1 = _cast(Z)
            res_data = f(X1.data, Y1.data, Z1.data)

            dims = mergedims(mergedims(X, Y), Z)
            coords = mergecoords(mergecoords(X, Y), Z)

            return XTensor(res_data, dims, coords)
        return wrapped
    return wrapper



import numpy as np
from numpy.typing import NDArray

from ._np import AxesLike, flatten, fold_indices, _axes_list



def argsmax(a: np.ndarray, axes: AxesLike=None) -> NDArray[np.int_]:
    b = flatten(a, axes)
    args = np.argmax(b, axis=-1)
    return fold_indices(args, [a.shape[i] for i in _axes_list(a, axes)])


def argsmin(a: np.ndarray, axes: AxesLike=None) -> NDArray[np.int_]:

    b = flatten(a, axes)
    args = np.argmin(b, axis=-1)

    return fold_indices(args, [a.shape[i] for i in _axes_list(a, axes)])


def nanargsmax(a: np.ndarray, axes: AxesLike=None) -> NDArray[np.int_]:

    b = flatten(a, axes)
    args = np.nanargmax(b, axis=-1)

    return fold_indices(args, [a.shape[i] for i in _axes_list(a, axes)])


def nanargsmin(a: np.ndarray, axes: AxesLike=None) -> NDArray[np.int_]:

    b = flatten(a, axes)
    args = np.nanargmin(b, axis=-1)

    return fold_indices(args, [a.shape[i] for i in _axes_list(a, axes)])



from __future__ import annotations

from typing import Any, Callable, Dict, Generic, List, Optional, Sequence, Tuple, TypeVar, overload
from numpy.typing import ArrayLike, NDArray
import numpy as np

from .broadcast import vanilla_broadcaster, Broadcaster
from .typing import BinaryOperator, DimLike, DimsLike, TensorLike, Array

from ._decors import promote_binary_operator

from .basic_utils._base import to_xtensor
from .basic_utils._misc import strip



def inject_broadcast(broadcaster: Broadcaster):
    promote = promote_binary_operator(broadcaster)
    def wrapper(f: Callable[[XTensor], BinaryOperator[np.ndarray]]
        ):
        def wrapped(self: XTensor, other: TensorLike, /) -> XTensor:
            binop = promote(f(self))
            if isinstance(other, XTensor):
                return binop(self, other)
            try:
                return binop(self, to_xtensor(other))
            except TypeError:
                return NotImplemented
        return wrapped
    return wrapper


class XTensor:
    def __init__(self, data: Array,
        dims: Optional[Sequence[str|None]]=None,
        coords: Optional[Sequence[Sequence[Any]|NDArray[Any]|None]]=None,
            ) -> None:


        self.data = data.__array__()


        self._dim_axis_dict: Dict[str,int] = dict()

        self._dims: List[str|None]
        self._coords: List[NDArray[Any]|None]

        self.set_dims(dims)
        self.set_coords(coords)

    def viewcopy(self) -> XTensor:
        return XTensor(self.data, self.dims, self.coords)
    
    def item(self) -> float:
        return self.data.item()

    @property
    def dims(self) -> Tuple[str|None,...]:
        return tuple(self._dims)

    @property
    def coords(self) -> Tuple[NDArray[Any]|None,...]:
        return tuple(self._coords)

    @property
    def shape(self) -> Tuple[int,...]:
        return self.data.shape

    @property
    def rank(self) -> int:
        return len(self.data.shape)

    def get_axis(self, dim: DimLike) -> int:
        if isinstance(dim, str): return self._dim_axis_dict[dim]
        if isinstance(dim, int):
            if dim < 0: dim = dim + self.rank
            if dim < 0 or dim >= self.rank: raise ValueError(f'Invalid axis: {dim}, rank={self.rank}')
            return dim
        
        try:
            assert isinstance(dim, tuple) 
            assert isinstance(dim[0], str)
            assert isinstance(dim[1], int)
            assert len(dim) == 2
        except AssertionError as e:
            raise ValueError('Dimlike should be int, str, or (str, int)') from e


        try:
            return self.get_axis(dim[0])
        except KeyError:
            return self.get_axis(dim[1])

    def get_axes(self, dims: DimLike|DimsLike|None) -> List[int]:
        if dims is None:
            return [i for i in range(self.rank)]

        if isinstance(dims, (int, str, tuple)): return [self.get_axis(dims)]
        axes = [self.get_axis(dim) for dim in dims]
        assert len(axes) == len(set(axes))
        return axes
            
    def set_dims(self, dims: Sequence[str|None]|None) -> None:
        if dims is None:
            self._dims = [None for _ in self.data.shape]
        else:
            assert len(dims) == len(self.data.shape)
            assert '' not in dims
            _dims_without_none = [dim for dim in dims if dim is not None]
            assert len(_dims_without_none) == len(set(_dims_without_none))
            self._dims = list(dims)

            for axis, dim in enumerate(dims):
                if dim is not None:
                    self._dim_axis_dict[dim] = axis

    def set_dim(self, dim: DimLike, newdim: str|None) -> None:
        axis = self.get_axis(dim)
        new_dims = self._dims.copy()
        new_dims[axis] = newdim
        self.set_dims(new_dims)

    def set_coords(self, coords: Sequence[Sequence[Any]|NDArray[Any]|None]|None) -> None:
        if coords is None:
            self._coords = [None for _ in self.data.shape]
        else:
            coords_clean: List[NDArray[Any]|None] = []
            assert len(coords) == len(self.data.shape)
            for axis, coord in enumerate(coords):
                if coord is not None:
                    assert len(coord) == self.data.shape[axis]
                    coords_clean.append(np.array(coords[axis]))
                else:
                    coords_clean.append(None)
            self._coords = coords_clean

    def set_coord(self, dim: DimLike, coord: Sequence[Any]|NDArray[Any]|None) -> None:
        axis = self.get_axis(dim)
        new_coords = self._coords.copy()
        new_coords[axis] = np.array(coord)
        self.set_coords(new_coords)


    def slc(self, dim: DimLike, slc: slice) -> XTensor:
        axis = self.get_axis(dim)

        slices: List[slice] = [slice(None, None, None)]*self.rank
        slices[axis] = slc

        data = self.data[tuple(slices)]
        coords = list(self.coords)
        
        coord = coords[axis]
        if coord is not None:
            coords[axis] = coord[slc]

        return XTensor(data, dims=self.dims, coords=coords)


    def get(self, dim: DimLike, index: int) -> XTensor:
        axis = self.get_axis(dim)
        
        slices: List[slice|int] = [slice(None, None, None)]*self.rank
        slices[axis] = index

        data = self.data[tuple(slices)]

        dims = strip(self.dims, [axis])
        coords = strip(self.coords, [axis])
        
        return XTensor(data, dims=dims, coords=coords)

    def __repr__(self):
        _repr = 'Tensor\n'
        _repr += f'shape={self.shape}\n'
        _repr += f'dims={self.dims}\n'

        _repr += f'coords:\n'
        for axis, coord in enumerate(self.coords):
            if coord is None:
                _repr += f'{axis}: None\n'
            if coord is not None:
                _repr += f'{axis}: {coord[0]}..{coord[-1]}\n'


        _repr += self.data.__repr__()

        return _repr

    def __neg__(self) -> XTensor:
        return XTensor(data=-self.data, dims=self._dims, coords=self._coords)

    @inject_broadcast(vanilla_broadcaster)
    def __add__(self): return lambda X, Y: X+Y

    @inject_broadcast(vanilla_broadcaster)
    def __radd__(self): return lambda X, Y: X+Y

    @inject_broadcast(vanilla_broadcaster)
    def __sub__(self): return lambda X, Y: X-Y

    @inject_broadcast(vanilla_broadcaster)
    def __rsub__(self): return lambda X, Y: Y-X

    @inject_broadcast(vanilla_broadcaster)
    def __mul__(self): return lambda X, Y: X*Y

    @inject_broadcast(vanilla_broadcaster)
    def __rmul__(self): return lambda X, Y: X*Y

    @inject_broadcast(vanilla_broadcaster)
    def __truediv__(self): return lambda X, Y: X/Y

    @inject_broadcast(vanilla_broadcaster)
    def __rtruediv__(self): return lambda X, Y: Y/X

    @inject_broadcast(vanilla_broadcaster)
    def __pow__(self): return lambda X, Y: X**Y

    @inject_broadcast(vanilla_broadcaster)
    def __rpow__(self): return lambda X, Y: Y**X

    @inject_broadcast(vanilla_broadcaster)
    def __eq__(self): return lambda X, Y: X==Y

    @inject_broadcast(vanilla_broadcaster)
    def __lt__(self): return lambda X, Y: X<Y

    @inject_broadcast(vanilla_broadcaster)
    def __gt__(self): return lambda X, Y: X>Y

    @inject_broadcast(vanilla_broadcaster)
    def __le__(self): return lambda X, Y: X<=Y

    @inject_broadcast(vanilla_broadcaster)
    def __ge__(self): return lambda X, Y: X>=Y

    def __array__(self): return self.data





from __future__ import annotations
from collections.abc import Sequence
from numbers import Real
import numpy as np
from typing import TYPE_CHECKING


if TYPE_CHECKING:
    from .._base import XTensor
    from ..typing import TensorLike


def to_xtensor(x: TensorLike) -> XTensor:
    from .._base import XTensor
    if isinstance(x, XTensor): return x
    
    try:
        return XTensor(x.__array__()) # type: ignore
    except AttributeError: pass

    if isinstance(x, Real):
        return XTensor(np.array(x))

    if isinstance(x, Sequence):
        return XTensor(np.array(list(x)))

    raise TypeError(f'Object cannot be safely converted to XTensor')



from __future__ import annotations

from typing import TYPE_CHECKING, List, Literal, Sequence, Tuple
import numpy as np
from ..typing import AxesPermutation, Dims, Coords

from ._generalize import generalize_at_0, generalize_at_1

from ._coords import coords_same


if TYPE_CHECKING: from .._base import XTensor


def permutation_well_defined(axes: AxesPermutation, rank: int|None=None) -> bool:
    axes_nnon = [axis for axis in axes if axis is not None]
    condition = (len(axes_nnon) == len(set(axes_nnon)))

    if rank is not None:
        condition = condition and all([axis in axes_nnon for axis in range(rank)])

    return condition


@generalize_at_0
def permute(X: XTensor, axes: AxesPermutation) -> XTensor:
    '''
    
    '''
    from .._base import XTensor

    data_ = X.data
    axes = axes.copy()
    newaxis = len(data_.shape)

    axes_refined: List[int] = []
    newdims = []
    newcoords = []
    
    for axis_t, axis_x in enumerate(axes):
        if axis_x is None:
            data_ = data_[...,None]
            axes_refined.append(newaxis)
            newdims.append(None)
            newcoords.append(None)
            newaxis += 1
        else:
            axes_refined.append(axis_x)
            newdims.append(X.dims[axis_x])
            newcoords.append(X.coords[axis_x])

    data_ = data_.transpose(*axes_refined)

    Y = XTensor(data_, dims=newdims, coords=newcoords)
    return Y


@generalize_at_0
def newdims(X: XTensor, *,
        dims: Dims|None=None, coords: Coords|None=None,
        position: Literal['left','right']='left') -> XTensor:
    '''
    Pad singleton dimensions to the given tensor with the specified dimension
    names and coordinates.
    ''' 
    from .._base import XTensor
    if dims is not None and coords is not None:
        assert len(dims) == len(coords)

        
    _x = X.data

    if dims is not None:
        n_dims = len(dims)

    else:
        if coords is None: raise ValueError('dims and coords cannot both be None')
        n_dims = len(coords)

    if dims is None:
        newdims = [None for _ in range(n_dims)]
    else:
        newdims = dims

    if coords is None:
        newcoords = [None for _ in range(n_dims)]
    else:
        newcoords = coords


    if position == 'left':
        _y = _x.reshape(*[1 for _ in range(n_dims)], *_x.shape)
        return XTensor(_y, dims=newdims+list(X.dims), coords=newcoords+list(X.coords))

    else:
        _y = _x.reshape(*_x.shape, *[1 for _ in range(n_dims)])
        return XTensor(_y, dims=list(X.dims)+newdims, coords=list(X.coords)+newcoords)


@generalize_at_0
@generalize_at_1
def align(X: XTensor, Y: XTensor) -> Tuple[XTensor, XTensor]:
    '''
    Pad singleton unnamed dimensions to (at most) one of the two tensors so that the
    returned tensors have the same number of dimnesions.
    '''
    if len(X.shape) > len(Y.shape):
        return X, newdims(Y, dims=[None for _ in range(len(Y.shape), len(X.shape))])
    
    if len(X.shape) < len(Y.shape):
        return newdims(X, dims=[None for _ in range(len(X.shape), len(Y.shape))]), Y

    return X, Y


def stack(x: Sequence[XTensor],
        newdim: str|None=None, 
        position: Literal['left', 'right']='left') -> XTensor:
    from .._base import XTensor

    for X in x:
        if not X.shape == x[0].shape:
            raise ValueError('All tensors should have the same shape for stacking')

        if not X.dims == x[0].dims:
            raise ValueError('All tensors should have the same dimension names for stacking')
    
        if not coords_same(list(X.coords), list(x[0].coords)):
            raise ValueError('All tensors should have the same coordinatees for stacking')

    axis = 0 if position == 'left' else -1
    data = np.stack([X.data for X in x], axis=axis)

    return XTensor(data, dims=[newdim]+list(x[0].dims), coords=[None]+list(x[0].coords))


def shapes_broadcastable(a: Sequence[int], b: Sequence[int]) -> bool:
    for sa, sb in zip(a[::-1], b[::-1]):
        if sa != sb and sa != 1 and sb != 1: return False
    return True



@generalize_at_0
def shape(X: XTensor): return X.shape


@generalize_at_0
def rank(X: XTensor): return X.rank




from __future__ import annotations

import numpy as np

from .. import tensor as xtt


@xtt.generalize_at_0
def diagonal(X: xtt.XTensor, /, dim1: xtt.DimLike, dim2: xtt.DimLike, dim_out: str|None) -> xtt.XTensor:
    '''
        Reduce a tensor x by taking the diagonal elements along [dim1] and
        [dim2]
    '''

    axis1 = X.get_axis(dim1)
    axis2 = X.get_axis(dim2)

    _y = np.diagonal(X.data, axis1=axis1, axis2=axis2)

    return xtt.XTensor(_y, 
            dims=xtt.strip(X.dims, [axis1, axis2]) + [dim_out],
            coords=xtt.strip(X.coords, [axis1, axis2]) + [None],
            )




from __future__ import annotations


from .. import tensor as xtt


class Functional:
    def __init__(self):
        self.name: str = 'NOTIMPLEMENTED_FUNCTIONAL'

    def __call__(self, x: xtt.TensorLike) -> xtt.XTensor:
        raise NotImplementedError


class Identity(Functional):
    def __init__(self, name='I'):
        self.name = name

    def __call__(self, x: xtt.TensorLike) -> xtt.XTensor:
        return xtt.to_xtensor(x)


class Pipe(Functional):
    '''
        Pipe(f1, f2, ..., fn)(x) = fn(...f2(f1(x))...)
    '''
    def __init__(self, *f: Functional, delim: str='.'):
        self.f = f
        self.delim = delim
        self.name = delim.join([_f.name for _f in self.f])

    def __call__(self, x: xtt.TensorLike) -> xtt.XTensor:
        _y = xtt.to_xtensor(x)
        for _f in self.f:
            _y = _f(_y)
        return _y



from __future__ import annotations
from abc import abstractmethod

'''
Template Broadcaster:

    First specify a template, then map axes of tensors onto the template

'''

from typing import TYPE_CHECKING, List, Literal, Optional, Sequence, Union
import numpy as np
from ._types import Broadcaster

if TYPE_CHECKING:
    from .._base import XTensor

from ..typing import Dims, Coords
from ..basic_utils import permute, permutation_well_defined, mergecoords, mergedims


Channel = Union[str, None]


class AxisSelector:
    def __init__(self) -> None:
        self._label: str
        self.channel: str|None

    @abstractmethod
    def select_axis(self, X: XTensor, channel: str|None=None) -> int|None:
        '''
        Select an axis from the tensor X
        '''

    @property
    def label(self) -> str:
        return self._label


class DimNameSelector(AxisSelector):
    def __init__(self, dimname: str, 
            required: bool=False, channel: Optional[str]=None
            ) -> None:
        self.dimname = dimname
        self.required = required
        self.channel = channel
        self._label = dimname

    def select_axis(self, X: XTensor, channel: str|None=None) -> int|None:
        if self.channel is not None and channel != self.channel:
            return None
        try:
            return X.get_axis(self.dimname)
        except KeyError as e:
            if self.required: raise e
            return None



class IndexSelector(AxisSelector):
    def __init__(self, axis: int, channel: Optional[str]=None) -> None:
        self.axis = axis
        self.channel = channel
        self._label = str(axis)


    def select_axis(self, X: XTensor, channel: str|None=None) -> int | None:
        if self.channel is not None and channel != self.channel:
            return None
        
        if self.axis >= 0:
            return self.axis
        else:
            return self.axis + X.rank


class Template:
    def __init__(self, *selectors: AxisSelector) -> None:

        try:
            for sel in selectors: assert isinstance(sel, AxisSelector)
        except AssertionError as e:
            raise TypeError('selectors must be instances of AxisSelector') from e

        self.selectors = selectors
        self._dims: Dims
        self._coords: Coords
        self.clear()
        self.check_well_defined()


    @classmethod
    def from_dims_channels(
            cls,
            dims: Sequence[Union[str, int]],
            channels: Sequence[str|None]|None=None) -> Template:
        selectors = []

        if channels is None: channels = [None for _ in range(len(dims))]

        assert len(dims) == len(channels)

        for d, c in zip(dims, channels):
            if isinstance(d, str):
                selectors.append(DimNameSelector(d, channel=c))
            else:
                selectors.append(IndexSelector(d, channel=c))
        
        return cls(*selectors)

    def check_well_defined(self):
        names = [sel.dimname for sel in self.selectors if isinstance(sel, DimNameSelector)]
        assert len(names) == len(set(names))

        axes = [sel.axis for sel in self.selectors if isinstance(sel, IndexSelector)]
        assert len(axes) == len(set(axes))

    def cast_and_update(self, X: XTensor, channel: str|None=None) -> XTensor:
        axesp = [sel.select_axis(X, channel) for sel in self.selectors]
        assert permutation_well_defined(axesp, X.rank), 'Casting impossible'

        X1 = permute(X, axesp)

        self._dims = mergedims(self._dims, X1)
        self._coords = mergecoords(self._coords, X1)

        return X1

    def clear(self):
        self._dims = [None for _ in range(len(self.selectors))]
        self._coords= [None for _ in range(len(self.selectors))]

    @property
    def dims(self) -> Dims:
        return self._dims

    @property
    def coords(self) -> Coords:
        return self._coords

    def __repr__(self) -> str:
        s = 'Dimension Template'
        s1 = 'dim.    '
        s2 = 'channel '

        for selector in self.selectors:
            label = selector.label
            channel = str(selector.channel)

            width = max(len(label), len(channel)) + 1

            s1 += f'{label:{width}}'
            s2 += f'{channel:{width}}'


        return s + '\n' + s1 + '\n' + s2

